from typing import Literal
from dataclasses import dataclass
from enum import Enum
from typing import Iterable, Generator
from PIL.Image import Image
from fitz import Document
from doc_page_extractor import clip, Rectangle, Layout, LayoutClass, OCRFragment, ExtractedResult
from .types import PDFPageExtractorProgressReport
from .document import DocumentExtractor, DocumentParams


class TextKind(Enum):
  TITLE = 0
  PLAIN_TEXT = 1
  ABANDON = 2

@dataclass
class Text:
  content: str
  rank: float
  rect: Rectangle

@dataclass
class BasicBlock:
  rect: Rectangle
  texts: list[Text]
  font_size: float

@dataclass
class TextBlock(BasicBlock):
  kind: TextKind
  has_paragraph_indentation: bool = False
  last_line_touch_end: bool = False

class AssetKind(Enum):
  FIGURE = 0
  TABLE = 1
  FORMULA = 2

@dataclass
class AssetBlock(BasicBlock):
  image: Image
  kind: AssetKind

Block = TextBlock | AssetBlock

class PDFPageExtractor:
  def __init__(
      self,
      device: Literal["cpu", "cuda"],
      model_dir_path: str,
      debug_dir_path: str | None = None,
    ):
    self._doc_extractor: DocumentExtractor = DocumentExtractor(
      device=device,
      model_dir_path=model_dir_path,
      debug_dir_path=debug_dir_path,
    )

  def extract(self, pdf: str | Document, report_progress: PDFPageExtractorProgressReport | None = None) -> Generator[Block, None, None]:
    for _, blocks, _ in self.extract_enumerated_blocks_and_image(
      pdf=pdf,
      report_progress=report_progress,
    ):
      yield from blocks

  def extract_enumerated_blocks_and_image(
      self,
      pdf: str | Document,
      page_indexes: Iterable[int] | None = None,
      report_progress: PDFPageExtractorProgressReport | None = None,

    ) -> Generator[tuple[int, list[Block], Image], None, None]:

    for page_index, result, layouts in self._doc_extractor.extract(DocumentParams(
      pdf=pdf,
      page_indexes=page_indexes,
      report_progress=report_progress,
    )):
      blocks = self._convert_to_blocks(result, layouts)
      page_range = self._texts_range(blocks)

      for block in blocks:
        if not isinstance(block, TextBlock) or \
           block.kind == TextKind.ABANDON:
          continue

        if len(block.texts) == 1:
          mean_line_height, x1, x2 = page_range
        else:
          mean_line_height, x1, x2 = self._texts_range((block,))

        first_text = block.texts[0]
        last_text = block.texts[-1]
        first_delta_x = (first_text.rect.lt[0] + first_text.rect.lb[0]) / 2 - x1
        last_delta_x = x2 - (last_text.rect.rt[0] + last_text.rect.rb[0]) / 2
        block.has_paragraph_indentation = first_delta_x > mean_line_height
        block.last_line_touch_end = last_delta_x < mean_line_height

      yield page_index, blocks, result.extracted_image

  def _convert_to_blocks(self, result: ExtractedResult, layouts: list[Layout]) -> list[Block]:
    store: list[tuple[Layout, Block]] = []
    def previous_block(cls: LayoutClass) -> Block | None:
      for i in range(len(store) - 1, -1, -1):
        layout, block = store[i]
        if cls == layout.cls:
          return block
        if cls != LayoutClass.ABANDON:
          return None
      return None

    for layout in layouts:
      cls = layout.cls
      if cls == LayoutClass.TITLE:
        store.append((layout, TextBlock(
          rect=layout.rect,
          kind=TextKind.TITLE,
          font_size=0.0,
          texts=self._convert_to_text(layout.fragments),
        )))
      elif cls == LayoutClass.PLAIN_TEXT:
        store.append((layout, TextBlock(
          rect=layout.rect,
          kind=TextKind.PLAIN_TEXT,
          font_size=0.0,
          texts=self._convert_to_text(layout.fragments),
        )))
      elif cls == LayoutClass.ABANDON:
        store.append((layout, TextBlock(
          rect=layout.rect,
          kind=TextKind.ABANDON,
          font_size=0.0,
          texts=self._convert_to_text(layout.fragments),
        )))
      elif cls == LayoutClass.FIGURE:
        store.append((layout, AssetBlock(
          rect=layout.rect,
          texts=[],
          kind=AssetKind.FIGURE,
          font_size=0.0,
          image=clip(result, layout),
        )))
      elif cls == LayoutClass.TABLE:
        store.append((layout, AssetBlock(
          rect=layout.rect,
          texts=[],
          kind=AssetKind.TABLE,
          font_size=0.0,
          image=clip(result, layout),
        )))
      elif cls == LayoutClass.ISOLATE_FORMULA:
        store.append((layout, AssetBlock(
          rect=layout.rect,
          texts=[],
          kind=AssetKind.FORMULA,
          font_size=0.0,
          image=clip(result, layout),
        )))
      elif cls == LayoutClass.FIGURE_CAPTION:
        block = previous_block(LayoutClass.FIGURE)
        if block is not None:
          assert isinstance(block, AssetBlock)
          block.texts.extend(self._convert_to_text(layout.fragments))
      elif cls == LayoutClass.TABLE_CAPTION or \
           cls == LayoutClass.TABLE_FOOTNOTE:
        block = previous_block(LayoutClass.TABLE)
        if block is not None:
          assert isinstance(block, AssetBlock)
          block.texts.extend(self._convert_to_text(layout.fragments))
      elif cls == LayoutClass.FORMULA_CAPTION:
        block = previous_block(LayoutClass.ISOLATE_FORMULA)
        if block is not None:
          assert isinstance(block, AssetBlock)
          block.texts.extend(self._convert_to_text(layout.fragments))

    self._fill_font_size_for_blocks(store)

    return [block for _, block in store]

  def _texts_range(self, blocks: Iterable[Block]) -> tuple[float, float, float]:
    sum_lines_height: float = 0.0
    texts_count: int = 0
    x1: float = float("inf")
    x2: float = float("-inf")

    for block in blocks:
      if not isinstance(block, TextBlock):
        continue
      if block.kind == TextKind.ABANDON:
        continue
      for text in block.texts:
        sum_lines_height += text.rect.size[1]
        texts_count += 1
        for x, _ in text.rect:
          x1 = min(x1, x)
          x2 = max(x2, x)

    if texts_count == 0:
      return 0.0, 0.0, 0.0
    return sum_lines_height / texts_count, x1, x2

  def _convert_to_text(self, fragments: list[OCRFragment]) -> list[Text]:
    return [
      Text(
        content=f.text,
        rank=f.rank,
        rect=f.rect,
      )
      for f in fragments
    ]

  def _fill_font_size_for_blocks(self, store: list[tuple[Layout, Block]]):
    font_sizes: list[float] = []

    for layout, _ in store:
      if len(layout.fragments) == 0:
        font_sizes.append(0.0)
      else:
        sum_height: float = 0.0
        for fragment in layout.fragments:
          sum_height += fragment.rect.size[1]

        # without considering vertical writing, the height of a line of text is proportional to the font size.
        font_size = sum_height / len(layout.fragments)
        font_sizes.append(font_size)

    max_font_size: float = 1.0
    min_font_size: float = 1.0

    if len(font_sizes) > 0:
      max_font_size = max(font_sizes)
      min_font_size = min(font_sizes)

    if max_font_size == min_font_size:
      for _, block in store:
        block.font_size = 0
    else:
      for font_size, (_, block) in zip(font_sizes, store):
        block.font_size = (font_size - min_font_size) / (max_font_size - min_font_size)



from typing import cast, Any
from importlib.resources import files
from jinja2 import Environment, Template
from xml.etree.ElementTree import tostring, Element
from pydantic import SecretStr
from tiktoken import get_encoding, Encoding
from langchain_openai import ChatOpenAI
from langchain_core.messages import SystemMessage, HumanMessage
from ..template import create_env


class LLM:
  def __init__(
      self,
      key: str,
      url: str,
      model: str,
      token_encoding: str,
    ):
    self._templates: dict[str, Template] = {}
    self._encoding: Encoding = get_encoding(token_encoding)
    self._model = ChatOpenAI(
      api_key=cast(SecretStr, key),
      base_url=url,
      model=model,
      temperature=0.7,
    )
    prompts_path = files("pdf_craft").joinpath("data/prompts")
    self._env: Environment = create_env(prompts_path)

  def request(self, template_name: str, xml_data: Element, params: dict[str, Any]) -> str:
    template = self._template(template_name)
    prompt = template.render(**params)
    data = tostring(xml_data, encoding="unicode")
    response = self._model([
      SystemMessage(content=prompt),
      HumanMessage(content=data)
    ])
    return response.content

  def prompt_tokens_count(self, template_name: str, params: dict[str, Any]) -> int:
    template = self._template(template_name)
    prompt = template.render(**params)
    return len(self._encoding.encode(prompt))

  def encode_tokens(self, text: str) -> list[int]:
    return self._encoding.encode(text)

  def decode_tokens(self, tokens: list[int]) -> str:
    return self._encoding.decode(tokens)

  def count_tokens_count(self, text: str) -> int:
    return len(self._encoding.encode(text))

  def _template(self, template_name: str) -> Template:
    template = self._templates.get(template_name, None)
    if template is None:
      template = self._env.get_template(template_name)
      self._templates[template_name] = template
    return template



from __future__ import annotations

import re
import os
import sys

from typing import Generator, Iterable
from xml.etree.ElementTree import fromstring, tostring, Element
from .splitter import Group, Segment
from .common import TextInfo


class ChunkFile:
  def __init__(self, output_dir_path: str):
    self._output_dir_path: str = output_dir_path
    self._files: list[tuple[int, int, str]] = list(self._search_chunk_file(output_dir_path))
    self._files.sort(key=lambda x: (x[0], x[1]))
    self._overlapped_files: dict[tuple[int, int], str] = {}
    self._min_page_index: int = sys.maxsize
    self._max_page_index: int = 0

  def __enter__(self) -> ChunkFile:
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    if exc_type is not None:
      return
    for file_name in self._overlapped_files.values():
      self._remove_file(file_name)
    for start, end, file_name in self._files:
      if start > self._max_page_index:
        self._remove_file(file_name)
      if end < self._min_page_index:
        self._remove_file(file_name)

  def filter_origin_files(self, origin_dir_path: str) -> Generator[tuple[int, int, Element], None, None]:
    for start, end, file_name in self._search_chunk_file(origin_dir_path):
      self._register_page_range(start, end)
      if self._overlap_files(start, end):
        file_path = os.path.join(origin_dir_path, file_name)
        with open(file_path, "r", encoding="utf-8") as file:
          yield start, end, fromstring(file.read())

  def _search_chunk_file(self, dir_path: str) -> Generator[tuple[int, int, str], None, None]:
    for file_name in os.listdir(dir_path):
      matches = re.match(r"^chunk_(\d+_\d+)\.xml$", file_name)
      if not matches:
        continue
      start, end = matches.group(1).split("_")
      yield int(start) - 1, int(end) - 1, file_name

  def filter_groups(self, groups: Iterable[Group]) -> Generator[tuple[int, int, Group], None, None]:
    for group in groups:
      start = min(t.page_index for t in self._search_text_infos(group))
      end = max(t.page_index for t in self._search_text_infos(group))
      self._register_page_range(start, end)

      if self._overlap_files(start, end):
        yield start, end, group

  def _search_text_infos(self, group: Group):
    for item in group.body:
      if isinstance(item, Segment):
        for text_info in item.text_infos:
          yield text_info
      elif isinstance(item, TextInfo):
        yield item

  def _register_page_range(self, start: int, end: int):
    self._min_page_index = min(self._min_page_index, start)
    self._max_page_index = max(self._max_page_index, end)

  # return False if matches origin chunk file
  def _overlap_files(self, start: int, end: int):
    overlap_files = list(self._search_overlap_files(start, end))
    overlap_files.sort(key=lambda x: (x[0], x[1]))

    if len(overlap_files) == 1:
      overlap_start, overlap_end, _ = overlap_files[0]
      if start == overlap_start and end == overlap_end:
        return False

    if len(overlap_files) > 0:
      self._overlapped_files[(start, end)] = overlap_files

    return True

  def _search_overlap_files(self, start: int, end: int):
    for file_start, file_end, file_name in self._files:
      if end < file_start:
        continue
      if start > file_end:
        continue
      yield file_start, file_end, file_name

  def atomic_write_chunk(self, start: int, end: int, chunk_xml: Element):
    overlap_files = self._overlapped_files.pop((start, end), None)
    if overlap_files is not None:
      for _, _, file_name in overlap_files:
        self._remove_file(file_name)

    file_name = f"chunk_{start + 1}_{end + 1}.xml"
    file_path = os.path.join(self._output_dir_path, file_name)
    try:
      with open(file_path, "w", encoding="utf-8") as file:
        content = tostring(chunk_xml, encoding="unicode")
        file.write(content)
        file.flush()
    except Exception as e:
      if os.path.exists(file_path):
        os.unlink(file_path)
      raise e

  def _remove_file(self, file_name: str):
    file_path = os.path.join(self._output_dir_path, file_name)
    if os.path.exists(file_path):
      os.remove(file_path)


import os
import re

from typing import Iterable, Generator
from xml.etree.ElementTree import fromstring, Element

def normalize_xml_text(xml_text: str) -> str:
  return re.sub(r"\s+", " ", xml_text).strip()

def read_xml_files(dir_path: str, enable_kinds: Iterable[str]) -> Generator[tuple[Element, str, str, int, int], None, None]:
  for file_name, kind, index1, index2 in read_files(dir_path, enable_kinds):
    file_path = os.path.join(dir_path, file_name)
    with open(file_path, "r", encoding="utf-8") as file:
      root = fromstring(file.read())
      yield root, file_name, kind, index1, index2

def search_xml_and_indexes(kind: str, dir_path: str) -> Generator[tuple[str, int, int], None, None]:
  for file_name in os.listdir(dir_path):
    matches = re.match(r"^[a-zA-Z]+_\d+(_\d+)?\.xml$", file_name)
    if not matches:
      continue
    file_kind: str
    index1: str
    index2: str
    cells = re.sub(r"\..*$", "", file_name).split("_")
    if len(cells) == 3:
      file_kind, index1, index2 = cells
    else:
      file_kind, index1 = cells
      index2 = index1
    if kind != file_kind:
      continue
    yield file_name, int(index1) - 1, int(index2) - 1

def read_files(dir_path: str, enable_kinds: Iterable[str]) -> Generator[tuple[str, str, int, int], None, None]:
  for file_name in os.listdir(dir_path):
    matches = re.match(r"^[a-zA-Z]+_\d+(_\d+)?\.xml$", file_name)
    if not matches:
      continue

    kind: str
    index1: str
    index2: str = "-1"
    cells = re.sub(r"\..*$", "", file_name).split("_")

    if len(cells) == 3:
      kind, index1, index2 = cells
    else:
      kind, index1 = cells

    if kind not in enable_kinds:
      continue

    yield file_name, kind, int(index1), int(index2)

def search_xml_children(parent: Element) -> Generator[Element, None, None]:
  for child in parent:
    yield child
    yield from search_xml_children(child)

def parse_page_indexes(element: Element) -> list[int]:
  idx = element.get("idx")
  if idx is None:
    return []
  else:
    return [int(i) - 1 for i in idx.split(",")]

def encode_response(response: str) -> Element:
  response = re.sub(r"^```XML", "", response)
  response = re.sub(r"```$", "", response)
  try:
    return fromstring(response.replace("&", "&amp;"))
  except Exception as e:
    print(response)
    raise e


from io import TextIOWrapper
from enum import IntEnum
from typing import Callable
from dataclasses import dataclass


@dataclass
class TextIncision(IntEnum):
  MUST_BE = 2
  MOST_LIKELY = 1
  IMPOSSIBLE = -1
  UNCERTAIN = 0

@dataclass
class TextInfo:
  page_index: int
  tokens: int
  start_incision: TextIncision
  end_incision: TextIncision

@dataclass
class PageInfo:
  page_index: int
  main: TextInfo
  citation: TextInfo | None
  file: Callable[[], TextIOWrapper]


from __future__ import annotations
from dataclasses import dataclass
from typing import Iterable, Generator
from math import floor
from .segment import Segment
from .stream import Stream
from ..common import TextInfo


@dataclass
class Group:
  head_remain_tokens: int
  tail_remain_tokens: int
  head: list[TextInfo | Segment]
  body: list[TextInfo | Segment]
  tail: list[TextInfo | Segment]

def group(
    items: Iterable[TextInfo | Segment],
    max_tokens: int,
    gap_rate: float,
    tail_rate: float,
  ) -> Generator[Group, None, None]:

  gap_max_tokens = floor(max_tokens * gap_rate)
  assert gap_max_tokens > 0

  curr_group: _Group = _Group(_Attributes(
    max_tokens=max_tokens,
    gap_max_tokens=gap_max_tokens,
    tail_rate=tail_rate,
  ))
  curr_group.head.seal()
  stream: Stream[_Item] = Stream(items)

  while True:
    item = stream.get()
    if item is not None:
      success = curr_group.append(item)
      if success:
        continue

    if curr_group.body.has_any:
      yield curr_group.report()
    if item is not None:
      stream.recover(item)
    for tail_item in reversed(list(curr_group.tail)):
      stream.recover(tail_item)

    if not stream.has_buffer and item is None:
      # next item never comes
      break
    curr_group = curr_group.next()

_Item = TextInfo | Segment

@dataclass
class _Attributes:
  max_tokens: int
  gap_max_tokens: float
  tail_rate: float

class _Group:
  def __init__(self, attr: _Attributes):
    self._attr: _Attributes = attr
    body_max_tokens = attr.max_tokens - attr.gap_max_tokens * 2
    assert body_max_tokens > 0

    # head and tail are passed to LLM as additional text
    # to let LLM understand the context of the body text.
    self.head: _Buffer = _Buffer(attr.gap_max_tokens)
    self.tail: _Buffer = _Buffer(attr.gap_max_tokens)
    self.body: _Buffer = _Buffer(body_max_tokens)

  def append(self, item: _Item) -> bool:
    success: bool = False
    for buffer in (self.head, self.body, self.tail):
      if buffer.is_sealed:
        continue
      if not buffer.can_append(item):
        buffer.seal()
        continue
      buffer.append(item)
      success = True
      break
    return success

  def next(self) -> _Group:
    next_group: _Group = _Group(self._attr)
    next_head = next_group.head
    for item in reversed([*self.head, *self.body]):
      if next_head.can_append(item):
        next_head.append(item)
      else:
        next_head.reverse().seal()
        break
    return next_group

  def report(self) -> Group:
    tokens: int = 0
    for buffer in (self.head, self.body, self.tail):
      tokens += buffer.tokens

    head_remain_tokens = self.head.tokens
    tail_remain_tokens = self.tail.tokens

    if tokens > self._attr.max_tokens:
      if self.body.tokens > self._attr.max_tokens:
        head_remain_tokens = 0
        tail_remain_tokens = 0
      else:
        tail_rate = self._attr.tail_rate
        remain_tokens = self._attr.max_tokens - self.body.tokens
        if self.head.tokens < remain_tokens * (1.0 - tail_rate):
          tail_remain_tokens = remain_tokens - self.head.tokens
        elif self.tail.tokens < remain_tokens * tail_rate:
          head_remain_tokens = remain_tokens - self.tail.tokens
        else:
          head_remain_tokens = round(remain_tokens * (1.0 - tail_rate))
          tail_remain_tokens = round(remain_tokens * tail_rate)

    head = list(self.head)
    tail = list(self.tail)

    if head_remain_tokens == 0:
      head = []
    if tail_remain_tokens == 0:
      tail = []

    return Group(
      head_remain_tokens=head_remain_tokens,
      tail_remain_tokens=tail_remain_tokens,
      head=head,
      body=list(self.body),
      tail=tail,
    )

class _Buffer:
  def __init__(self, max_tokens: int):
    self._max_tokens: int = max_tokens
    self._items: list[_Item] = []
    self._tokens: int = 0
    self._is_sealed: bool = False

  @property
  def is_sealed(self) -> bool:
    return self._is_sealed

  @property
  def has_any(self) -> bool:
    return len(self._items) > 0

  @property
  def tokens(self) -> int:
    return self._tokens

  def seal(self):
    self._is_sealed = True

  def reverse(self) -> _Buffer:
    self._items.reverse()
    return self

  def __iter__(self):
    return iter(self._items)

  def append(self, item: _Item):
    self._items.append(item)
    self._tokens += item.tokens

  def can_append(self, item: _Item) -> bool:
    if self._is_sealed:
      return False
    if len(self._items) == 0:
      return True
    next_tokens = self._tokens + item.tokens
    return next_tokens <= self._max_tokens


import os
import shutil

from json import dumps, loads
from typing import Iterable, Callable
from xml.etree.ElementTree import tostring, fromstring, Element
from ..pdf import PDFPageExtractor
from .llm import LLM
from .common import PageInfo, TextInfo, TextIncision
from .chunk_file import ChunkFile
from .ocr_extractor import extract_ocr_page_xmls
from .page import analyse_page
from .index import analyse_index, Index
from .citation import analyse_citations
from .main_text import analyse_main_texts
from .position import analyse_position
from .meta import extract_meta
from .chapter import generate_chapters
from .asset_matcher import ASSET_TAGS
from .types import AnalysingStep, AnalysingProgressReport, AnalysingStepReport
from .utils import search_xml_and_indexes


def analyse(
  llm: LLM,
  pdf_page_extractor: PDFPageExtractor,
  pdf_path: str,
  analysing_dir_path: str,
  output_dir_path: str,
  report_step: AnalysingStepReport | None = None,
  report_progress: AnalysingProgressReport | None = None,
):
  state_machine = _StateMachine(
    llm=llm,
    pdf_page_extractor=pdf_page_extractor,
    pdf_path=pdf_path,
    analysing_dir_path=analysing_dir_path,
    output_dir_path=output_dir_path,
    report_progress=report_progress,
    report_step=report_step,
  )
  state_machine.start()

_IndexXML = tuple[str, int, int]
_MARK_FILE_NAME = "MARK_DONE"

class _StateMachine:
  def __init__(
      self,
      llm: LLM,
      pdf_page_extractor: PDFPageExtractor,
      pdf_path: str,
      analysing_dir_path: str,
      output_dir_path: str,
      report_step: AnalysingStepReport | None,
      report_progress: AnalysingProgressReport | None,
    ):
    self._llm: LLM = llm
    self._pdf_page_extractor: PDFPageExtractor = pdf_page_extractor
    self._pdf_path: str = pdf_path
    self._analysing_dir_path: str = analysing_dir_path
    self._output_dir_path: str = output_dir_path
    self._report_step: AnalysingStepReport | None = report_step
    self._report_progress: AnalysingProgressReport | None = report_progress
    self._index: Index | None = None
    self._index_did_load: bool = False
    self._pages: list[PageInfo] | None = None

  def start(self):
    self._run_analyse_step("ocr", self._extract_ocr)
    self._run_analyse_step("pages", self._analyse_pages)
    self._run_analyse_step("index", self._analyse_index)
    self._run_analyse_step("citations", self._analyse_citations)
    self._run_analyse_step("main_texts", self._analyse_main_texts)
    self._run_analyse_step("position", self._analyse_position)
    self._run_analyse_step("meta", self._extract_meta)
    self._generate_chapters()

  def _run_analyse_step(self, dir_name: str, func: Callable[[str], None]):
    mark_path = os.path.join(self._analysing_dir_path, dir_name, _MARK_FILE_NAME)
    if not os.path.exists(mark_path):
      dir_path = self._ensure_dir_path(
        dir_path=os.path.join(self._analysing_dir_path, dir_name),
      )
      func(dir_path)
      self._atomic_write(mark_path, "")

  def _extract_ocr(self, dir_path: str):
    assets_path = self._ensure_dir_path(os.path.join(self._analysing_dir_path, "assets"))
    index_xmls = self._list_index_xmls("page", dir_path)

    for page_index, page_xml in extract_ocr_page_xmls(
      extractor=self._pdf_page_extractor,
      pdf_path=self._pdf_path,
      expected_page_indexes=set(i for _, i, _ in index_xmls),
      cover_path=os.path.join(dir_path, "cover.png"),
      assets_dir_path=assets_path,
      report_step=self._report_step,
      report_progress=self._report_progress,
    ):
      self._atomic_write(
        file_path=os.path.join(dir_path, self._xml_name("page", page_index)),
        content=tostring(page_xml, encoding="unicode"),
      )

  def _analyse_pages(self, dir_path: str):
    from_path = os.path.join(self._analysing_dir_path, "ocr")
    done_page_indexes: set[int] = set()
    done_page_names: dict[int, str] = {}

    for file_name, page_index, _ in search_xml_and_indexes("page", dir_path):
      done_page_indexes.add(page_index)
      done_page_names[page_index] = file_name

    index_xmls = self._list_index_xmls("page", from_path)

    if self._report_step is not None:
      self._report_step(AnalysingStep.ANALYSE_PAGE, len(index_xmls))

    for i, (raw_name, page_index, _) in enumerate(index_xmls):
      if page_index not in done_page_indexes:
        raw_page_xml = self._read_xml(os.path.join(from_path, raw_name))
        previous_response_xml: Element | None = None
        if page_index > 0:
          file_name = done_page_names[page_index - 1]
          file_path = os.path.join(dir_path, file_name)
          previous_response_xml = self._read_xml(file_path)

        response_xml = analyse_page(
          llm=self._llm,
          raw_page_xml=raw_page_xml,
          previous_page_xml=previous_response_xml,
        )
        self._atomic_write(
          file_path=os.path.join(dir_path, raw_name),
          content=tostring(response_xml, encoding="unicode"),
        )
        done_page_names[page_index] = f"page_{page_index + 1}.xml"

      if self._report_progress is not None:
        self._report_progress(i + 1)

  def _analyse_index(self, dir_path: str):
    if self._report_step is not None:
      self._report_step(AnalysingStep.EXTRACT_INDEX, 0)

    from_path = os.path.join(self._analysing_dir_path, "pages")
    json_index, index = analyse_index(
      llm=self._llm,
      raw=(
        (i, self._read_xml(os.path.join(from_path, file_name)))
        for file_name, i, _ in self._list_index_xmls("page", from_path)
      )
    )
    if json_index is not None:
      self._atomic_write(
        file_path=os.path.join(dir_path, "index.json"),
        content=dumps(
          obj=json_index,
          ensure_ascii=False,
          indent=2,
        ),
      )
    if index is not None:
      self._index = index
      self._index_did_load = True

  def _analyse_citations(self, dir_path: str):
    with ChunkFile(dir_path) as file:
      analyse_citations(
        llm=self._llm,
        file=file,
        pages=self._load_pages(),
        request_max_tokens=8000,
        tail_rate=0.15,
        report_step=self._report_step,
        report_progress=self._report_progress,
      )

  def _analyse_main_texts(self, dir_path: str):
    citations_dir_path = os.path.join(self._analysing_dir_path, "citations")
    with ChunkFile(dir_path) as file:
      analyse_main_texts(
        llm=self._llm,
        file=file,
        index=self._load_index(),
        pages=self._load_pages(),
        citations_dir_path=citations_dir_path,
        request_max_tokens=10000,
        gap_rate=0.1,
        report_step=self._report_step,
        report_progress=self._report_progress,
      )

  def _analyse_position(self, dir_path: str):
    main_texts_path = os.path.join(self._analysing_dir_path, "main_texts")
    with ChunkFile(dir_path) as file:
      chunk_xmls = file.filter_origin_files(main_texts_path)
      if self._report_step is not None:
        chunk_xmls = list(chunk_xmls)
        self._report_step(AnalysingStep.MARK_POSITION, len(chunk_xmls))

      for i, (start_idx, end_idx, chunk_xml) in enumerate(chunk_xmls):
        position_xml = analyse_position(self._llm, self._load_index(), chunk_xml)
        file.atomic_write_chunk(start_idx, end_idx, position_xml)
        if self._report_progress is not None:
          self._report_progress(i + 1)

  def _extract_meta(self, dir_path: str):
    page_xmls: list[Element] = []
    page_dir_path = os.path.join(self._analysing_dir_path, "pages")
    head_count = 5

    if self._report_step is not None:
      self._report_step(AnalysingStep.ANALYSE_META, 0)

    for file_name, page_index, _ in search_xml_and_indexes("page", page_dir_path):
      if page_index >= head_count:
        break
      file_path = os.path.join(page_dir_path, file_name)
      page_xml = self._read_xml(file_path)
      page_xmls.append(page_xml)

    meta_json = extract_meta(self._llm, page_xmls)
    if meta_json is not None:
      meta_file_path = os.path.join(dir_path, "meta.json")
      self._atomic_write(
        file_path=meta_file_path,
        content=dumps(meta_json, ensure_ascii=False, indent=2),
      )

  def _generate_chapters(self):
    if self._report_step is not None:
      self._report_step(AnalysingStep.GENERATE_CHAPTERS, 0)

    if self._index is not None:
      file_path = os.path.join(self._output_dir_path, "index.json")
      self._atomic_write(
        file_path=file_path,
        content=dumps(self._index.json, ensure_ascii=False),
      )
    asset_hash_set: set[str] = set()

    for id, chapter_xml in generate_chapters(
      llm=self._llm,
      chunks_path=os.path.join(self._analysing_dir_path, "position"),
    ):
      if id is None:
        file_name = "chapter.xml"
      else:
        file_name = f"chapter_{id}.xml"

      self._atomic_write(
        file_path=os.path.join(self._output_dir_path, file_name),
        content=tostring(chapter_xml, encoding="unicode"),
      )
      content_xml = chapter_xml.find("content")
      for child in content_xml:
        if child.tag in ASSET_TAGS:
          hash = child.get("hash", None)
          if hash is not None:
            asset_hash_set.add(hash)

    self._copy_file(
      src_path=os.path.join(self._analysing_dir_path, "ocr", "cover.png"),
      dst_path=os.path.join(self._output_dir_path, "cover.png"),
    )
    self._copy_file(
      src_path=os.path.join(self._analysing_dir_path, "index", "index.json"),
      dst_path=os.path.join(self._output_dir_path, "index.json"),
    )
    self._copy_file(
      src_path=os.path.join(self._analysing_dir_path, "meta", "meta.json"),
      dst_path=os.path.join(self._output_dir_path, "meta.json"),
    )

    if len(asset_hash_set) > 0:
      asset_path = self._ensure_dir_path(os.path.join(self._output_dir_path, "assets"))
      for hash in asset_hash_set:
        src_path = os.path.join(self._analysing_dir_path, "assets", f"{hash}.png")
        dst_path = os.path.join(asset_path, f"{hash}.png")
        if os.path.exists(src_path):
          shutil.copy(src_path, dst_path)

  def _ensure_dir_path(self, dir_path: str) -> str:
    os.makedirs(dir_path, exist_ok=True)
    return dir_path

  def _load_index(self) -> Index | None:
    if not self._index_did_load:
      index_file_path = os.path.join(self._analysing_dir_path, "index", "index.json")
      self._index_did_load = True

      if os.path.exists(index_file_path):
        with open(index_file_path, "r", encoding="utf-8") as file:
          self._index = Index(loads(file.read()))

    return self._index

  def _list_index_xmls(self, kind: str, dir_path: str) -> list[_IndexXML]:
    index_xmls = list(search_xml_and_indexes(kind, dir_path))
    index_xmls.sort(key=lambda x: x[1])
    return index_xmls

  def _load_pages(self) -> list[PageInfo]:
    if self._pages is None:
      pages: list[PageInfo] = []
      pages_path = os.path.join(self._analysing_dir_path, "pages")

      for file_name, page_index, _ in self._list_index_xmls("page", pages_path):
        file_path = os.path.join(pages_path, file_name)
        page_xml = self._read_xml(file_path)
        page = self._parse_page_info(file_path, page_index, page_xml)
        pages.append(page)

      pages.sort(key=lambda p: p.page_index)
      self._pages = pages

    return self._pages

  def _parse_page_info(self, file_path: str, page_index: int, root: Element) -> PageInfo:
    main_children: list[Element] = []
    citation: TextInfo | None = None

    for child in root:
      if child.tag == "citation":
        citation = self._parse_text_info(page_index, child)
      else:
        main_children.append(child)

    return PageInfo(
      page_index=page_index,
      citation=citation,
      file=lambda: open(file_path, "rb"),
      main=self._parse_text_info(page_index, main_children),
    )

  def _parse_text_info(self, page_index: int, children: Iterable[Element]) -> TextInfo:
    # When no text is found on this page, it means it is full of tables or
    # it is a blank page. We cannot tell if there is a cut in the context.
    start_incision: TextIncision = TextIncision.UNCERTAIN
    end_incision: TextIncision = TextIncision.UNCERTAIN
    first: Element | None = None
    last: Element | None = None

    for child in children:
      if first is None:
        first = child
      last = child

    if first is not None and last is not None:
      if first.tag == "text":
        start_incision = self._attr_value_to_kind(first.attrib.get("start-incision"))
      if last.tag == "text":
        end_incision = self._attr_value_to_kind(last.attrib.get("end-incision"))

    tokens = self._count_elements_tokens(children)

    return TextInfo(
      page_index=page_index,
      tokens=tokens,
      start_incision=start_incision,
      end_incision=end_incision,
    )

  def _count_elements_tokens(self, elements: Iterable[Element]) -> int:
    root = Element("page")
    root.extend(elements)
    xml_content = tostring(root, encoding="unicode")
    return self._llm.count_tokens_count(xml_content)

  def _attr_value_to_kind(self, value: str | None) -> TextIncision:
    if value == "must-be":
      return TextIncision.MUST_BE
    elif value == "most-likely":
      return TextIncision.MOST_LIKELY
    elif value == "impossible":
      return TextIncision.IMPOSSIBLE
    elif value == "uncertain":
      return TextIncision.UNCERTAIN
    else:
      return TextIncision.UNCERTAIN

  def _xml_name(self, kind: str, page_index: int, page_index2: int | None = None) -> str:
    if page_index2 is None:
      return f"{kind}_{page_index + 1}.xml"
    else:
      return f"{kind}_{page_index + 1}_{page_index2 + 1}.xml"

  def _read_xml(self, file_path: str) -> Element:
    with open(file_path, "r", encoding="utf-8") as file:
      return fromstring(file.read())

  def _copy_file(self, src_path: str, dst_path: str):
    if os.path.exists(src_path):
      shutil.copy(src_path, dst_path)

  def _atomic_write(self, file_path: str, content: str):
    try:
      with open(file_path, "w", encoding="utf-8") as file:
        file.write(content)
        file.flush()
    except Exception as e:
      if os.path.exists(file_path):
        os.unlink(file_path)
      raise e



from typing import Generator, Callable
from dataclasses import dataclass
from xml.etree.ElementTree import tostring, Element
from .segment import Segment
from .group import Group
from ..common import TextInfo
from ..llm import LLM


@dataclass
class PageXML:
  page_index: int
  xml: Element
  is_gap: bool = False

def get_and_clip_pages(llm: LLM, group: Group, get_element: Callable[[int], Element]) -> list[PageXML]:
  head = _get_pages(
    llm=llm,
    items=group.head,
    remain_tokens=group.head_remain_tokens,
    clip_tail=False,
    get_element=get_element,
  )
  tail = _get_pages(
    llm=llm,
    items=group.tail,
    remain_tokens=group.tail_remain_tokens,
    clip_tail=True,
    get_element=get_element,
  )
  body = _get_pages(
    llm=llm,
    items=group.body,
    remain_tokens=None,
    clip_tail=True,
    get_element=get_element,
  )
  page_xml_list: list[PageXML] = []

  for page_xml in reversed(list(head)):
    page_xml.is_gap = True
    page_xml_list.append(page_xml)

  for page_xml in body:
    page_xml_list.append(page_xml)

  for page_xml in tail:
    page_xml.is_gap = True
    page_xml_list.append(page_xml)

  return page_xml_list

def _get_pages(
    llm: LLM,
    items: list[TextInfo | Segment],
    remain_tokens: int | None,
    clip_tail: bool,
    get_element: Callable[[int], str],
  ) -> Generator[PageXML, None, None]:

  if len(items) == 0:
    return

  if remain_tokens is not None:
    tokens = sum(item.tokens for item in items)
    if remain_tokens == tokens:
      remain_tokens = None

  if remain_tokens is not None:
    assert len(items) == 1

  if remain_tokens is None:
    for item in items:
      if isinstance(item, TextInfo):
        yield PageXML(
          page_index=item.page_index,
          xml=get_element(item.page_index),
        )
      elif isinstance(item, Segment):
        for text_info in item.text_infos:
          yield PageXML(
            page_index=text_info.page_index,
            xml=get_element(text_info.page_index),
          )
  else:
    item = items[0]
    if isinstance(item, TextInfo):
      page_xml = get_element(item.page_index)
      page_xml = _clip_element(llm, page_xml, remain_tokens, clip_tail)
      if page_xml is not None:
        yield PageXML(
          page_index=item.page_index,
          xml=page_xml,
        )

    elif isinstance(item, Segment):
      text_infos, remain_tokens = _clip_segment(item, remain_tokens, clip_tail)
      page_xml_list: list[PageXML] = []
      for i, text_info in enumerate(text_infos):
        page_xml: Element | None = get_element(text_info.page_index)
        if (clip_tail and i == len(text_infos) - 1) or \
           (not clip_tail and i == 0):
          page_xml = _clip_element(llm, page_xml, remain_tokens, clip_tail)
        if page_xml is not None:
          page_xml_list.append(PageXML(
            page_index=text_info.page_index,
            xml=page_xml,
          ))
      if not clip_tail:
        page_xml_list.reverse()
      yield from page_xml_list

def _clip_segment(segment: Segment, remain_tokens: int, clip_tail: bool):
  clipped: list[TextInfo] = []
  iter_source = segment.text_infos
  if not clip_tail:
    iter_source = reversed(iter_source)

  for text_info in iter_source:
    clipped.append(text_info)
    if remain_tokens >= text_info.tokens:
      remain_tokens -= text_info.tokens
    else:
      break
  if not clip_tail:
    clipped.reverse()

  return clipped, remain_tokens

def _clip_element(llm: LLM, element: Element, remain_tokens: int, clip_tail: bool) -> Element | None:
  clipped_element = Element(element.tag, element.attrib)
  children: list[tuple[Element, int]] = []
  remain_tokens -= llm.count_tokens_count(tostring(clipped_element, encoding="unicode"))

  for child in element:
    child_text = tostring(child, encoding="unicode")
    child_tokens = llm.count_tokens_count(child_text)
    children.append((child, child_tokens))
  if not clip_tail:
    children.reverse()

  if len(children) == 0:
    text = element.text
    if text is None:
      return None
    tokens = llm.encode_tokens(text)
    if clip_tail:
      tokens = tokens[:remain_tokens]
    else:
      tokens = tokens[len(tokens) - remain_tokens:]
    if len(tokens) == 0:
      return None
    clipped_element.text = llm.decode_tokens(tokens)

  else:
    clipped_children: list[Element] = []
    for child, tokens_count in children:
      if remain_tokens >= tokens_count:
        remain_tokens -= tokens_count
        clipped_children.append(child)
      else:
        child = _clip_element(llm, child, remain_tokens, clip_tail)
        if child is not None:
          clipped_children.append(child)
        break
    if len(clipped_children) == 0:
      return None
    if not clip_tail:
      clipped_children.reverse()
    clipped_element.extend(clipped_children)

  return clipped_element


from setuptools import setup, find_packages

setup(
  name="pdf-craft",
  version="0.0.6",
  author="Tao Zeyu",
  author_email="i@taozeyu.com",
  url="https://github.com/oomol-lab/pdf-craft",
  description="PDF craft can convert PDF files into various other formats. This project will focus on processing PDF files of scanned books. The project has just started.",
  packages=find_packages(),
  long_description=open("./README.md", encoding="utf8").read(),
  long_description_content_type="text/markdown",
  include_package_data=True,
  package_data={
    "pdf_craft": ["data/**/*.jinja"],
  },
  install_requires=[
    "jinja2>=3.1.5,<4.0.0",
    "pyMuPDF>=1.25.3,<2.0.0",
    "alphabet-detector>=0.0.7,<1.0.0",
    "shapely>=2.0.6,<3.0.0",
    "tiktoken>=0.9.0,<1.0.0",
    "doc-page-extractor==0.0.6",
    "langchain>=0.3.17,<0.4.0",
    "langchain_community>=0.3.16,<0.4.0",
    "langchain_core>=0.3.35,<0.4.0",
    "langchain_openai>=0.3.6,<0.4.0",
    "langchain_anthropic>=0.3.7,<0.4.0",
  ],
)


from __future__ import annotations
from dataclasses import dataclass
from typing import Iterable, Generator
from .stream import Stream
from ..common import TextIncision, TextInfo


_MIN_LEVEL = -1

@dataclass
class Segment:
  tokens: int
  text_infos: list[TextInfo]

def allocate_segments(text_infos: Iterable[TextInfo], max_tokens: int) -> Generator[TextInfo | Segment, None, None]:
  segment = _collect_segment(Stream(text_infos), _MIN_LEVEL)
  for item in segment.children:
    if isinstance(item, _Segment):
      for segment in _split_segment_if_need(item, max_tokens):
        yield _transform_segment(segment)
    elif isinstance(item, TextInfo):
      yield item

def _transform_segment(segment: _Segment):
  children = list(_deep_iter_segment(segment))
  if len(children) == 1:
    return children[0]
  else:
    return Segment(
      tokens=segment.tokens,
      text_infos=children,
    )

@dataclass
class _Segment:
  level: int
  tokens: int
  start_incision: TextIncision
  end_incision: TextIncision
  children: list[TextInfo | _Segment]

def _collect_segment(stream: Stream[TextInfo], level: int) -> _Segment:
  start_incision: TextIncision = TextIncision.IMPOSSIBLE
  end_incision: TextIncision = TextIncision.IMPOSSIBLE
  children: list[TextInfo | _Segment] = []

  while True:
    text = stream.get()
    if text is None:
      break
    if len(children) == 0: # is the first
      start_incision = text.start_incision
      children.append(text)
    else:
      pre_text = children[-1]
      incision_level = _to_level(
        pre_text.end_incision,
        text.start_incision,
      )
      if incision_level < level:
        stream.recover(text)
        end_incision = text.end_incision
        break
      elif incision_level > level:
        stream.recover(text)
        stream.recover(pre_text)
        segment = _collect_segment(stream, incision_level)
        children[-1] = segment
      else:
        children.append(text)

  tokens: int = 0
  for child in children:
    tokens += child.tokens

  return _Segment(
    level=level,
    tokens=tokens,
    start_incision=start_incision,
    end_incision=end_incision,
    children=children,
  )

def _split_segment_if_need(segment: _Segment, max_tokens: int):
  if segment.tokens <= max_tokens:
    yield segment
  else:
    tokens: int = 0
    children: list[TextInfo | _Segment] = []

    for item in _unfold_segments(segment, max_tokens):
      if len(children) > 0 and tokens + item.tokens > max_tokens:
        yield _create_segment(tokens, children, segment.level)
        tokens = 0
        children = []
      tokens += item.tokens
      children.append(item)

    if len(children) > 0:
      yield _create_segment(tokens, children, segment.level)

def _unfold_segments(segment: _Segment, max_tokens: int) -> Generator[TextInfo | _Segment]:
  for item in segment.children:
    if item.tokens > max_tokens and isinstance(item, _Segment):
      for sub_item in _split_segment_if_need(item, max_tokens):
        yield sub_item
    else:
      yield item

def _create_segment(tokens: int, children: list[TextInfo | _Segment], level: int) -> _Segment:
  return _Segment(
    level=level,
    tokens=tokens,
    children=children,
    start_incision=children[0].start_incision,
    end_incision=children[-1].end_incision,
  )

def _deep_iter_segment(segment: _Segment) -> Generator[TextInfo, None, None]:
  for child in segment.children:
    if isinstance(child, _Segment):
      yield from _deep_iter_segment(child)
    elif isinstance(child, TextInfo):
      yield child

def _to_level(left_incision: TextIncision, right_incision: TextIncision) -> int:
  return max(_MIN_LEVEL, left_incision.value + right_incision.value)



from typing import Iterable, Iterator, TypeVar, Generic


E = TypeVar("E")

class Stream(Generic[E]):
  def __init__(self, elements: Iterable[E]):
    self._iterator: Iterator[E] = iter(elements)
    self._buffer: list[E] = []

  @property
  def has_buffer(self) -> bool:
    return len(self._buffer) > 0

  def recover(self, element: E):
    self._buffer.append(element)

  def get(self) -> E | None:
    if len(self._buffer) > 0:
      return self._buffer.pop()
    else:
      return next(self._iterator, None)